% Hierarchical construction of a hierarchical inverse.
% A quadtree is used, and boxes are merged by pairs or two.
% Dense linear algebra is used throughout, so this is O(N^1.5)
% for an n x n grid with N = n^2 nodes.
%
% The figure illustrates the ordering of points in the grid
% for the case nside=5 
%
%        5-----10----15----20----25
%        |     |     |     |     | 
%   n    4-----9-----14----19----24
%   s    |     |     |     |     | 
%   i    3-----8-----13----18----23
%   d    |     |     |     |     | 
%   e    2-----7-----12----17----22
%        |     |     |     |     | 
%        1-----6-----11----16----21
%
%                 nside
%
% Note that the function that constructs the stiffness matrix has several
% different calling options:
%
%   flag_matrix='orig'   The usual 5-point discrete Laplacian 4/-1/-1/-1/-1
%
%   flag_matrix='rcon'   A grid Laplacian where the conductivities of all
%                        bars are drawn at random.
%
%   flag_matrix='rcut'   A grid Laplacian for a grid where some bars have
%                        been cut at random.
%
%   flag_matrix='peri'   Conductivities vary periodically.
%
%   flag_matrix='crac'   Two long cracks in the grid.

function main01()

rng('default')
rng(0)

nlevel = 3;
nleaf  = 5;
nside  = nleaf*(2^nlevel);
N      = nside*nside;
fprintf(1,'\n=== Parameters: nlevel=%2d   nleaf=%3d   nside=%4d   ntot=%d\n',...
        nlevel,nleaf,nside,N)

%%% Construct the stiffness matrix (this is a sparse matrix).
flag_matrix = 'rcon';
flag_anchor = 1;
A = LOCAL_get_A(nside,nside,flag_matrix,flag_anchor);

%%% Perform the actual nested dissection process.
fprintf(1,'\n=== Construct A inverse\n')
tic
NODES = LOCAL_get_Ainv_hierarchical(A,nlevel,nleaf);
fprintf(1,'    Time   required for full inversion = %0.2f (sec)\n',toc)
fprintf(1,'    Memory required for full inversion = %d (KB)\n',...
               round(LOCAL_get_bytes(NODES)/1024))

%%% Extract the Schur complement for the top node in "normal" matrix format.
S = NODES{15,1};

%%% Verify that the Schur complement is correct.
[J_rim,J_int] = LOCAL_get_Jrim_Jint_for_square(nside);
if (N < 5000) % Directly compute the "exact" Schur complement and compare them:
  fprintf(1,'\n=== Verifying the solution by computing the full Schur complement\n')
  S_ref = full(A(J_rim,J_rim) - A(J_rim,J_int)*(A(J_int,J_int)\A(J_int,J_rim)));
  fprintf(1,'    Absolute error in "max-norm" = %9.2e\n',norm(S - S_ref,Inf))
else
  fprintf(1,'\n=== Verifying the solution for a problem with boundary load only\n')
  f             = zeros(N,1);
  f(J_rim)      = randn(4*nside-4,1);
  u_ref         = A\f;
  u             = S\f(J_rim);
  fprintf(1,'    Absolute error from a test solve = %9.2e\n',norm(u - u_ref(J_rim)))
  fprintf(1,'    Relative error from a test solve = %9.2e\n',norm(u - u_ref(J_rim))/norm(u_ref(J_rim)))
end

%%% Verify the solution for a full solve (non-zero load in the interior).
fprintf(1,'\n=== Verifying the solution for a full domain solve\n')
f     = randn(N,1);
tic;
u     = LOCAL_apply_Ainv(NODES,f);
t_app = toc;
u_ref = A\f;
fprintf(1,'    Time required for application    = %0.2f (sec)\n',t_app)
fprintf(1,'    Absolute error from a test solve = %9.2e\n',norm(u - u_ref))
fprintf(1,'    Relative error from a test solve = %9.2e\n',norm(u - u_ref)/norm(u_ref))
fprintf(1,'\n')

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is the key function in this program. 
% For documention on the call see the top of the file.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function A = LOCAL_get_A(nside1,nside2,flag_matrix,flag_anchor)

if (nargin < 4)
  flag_anchor = 0;
end

if (nargin == 1)
  nside2 = nside1;
end

BARS  = LOCAL_get_links(nside1,nside2);
nbars = size(BARS,2);

if strcmp(flag_matrix,'orig')
    
  BARS(3,:) = ones(1,nbars);
  
elseif strcmp(flag_matrix,'rcon')
    
  BARS(3,:) = 1 + rand(1,nbars);
  
elseif strcmp(flag_matrix,'rcut')
    
  p = 0.04;
  BARS(3,:) = (rand(1,nbars) > p);
  
elseif strcmp(flag_matrix,'peri')
  
  [XX1,XX2] = meshgrid(0:(nside1-1),0:(nside2-1));
  xx        = [reshape(XX1,1,nside1*nside2);reshape(XX2,1,nside1*nside2)];
  h         = 1/(max([nside1,nside2])-1);
  Ncell     = round(min(nside1,nside2)/25);
  XX1sc     = h*xx(1,BARS(1,:));
  XX2sc     = h*xx(2,BARS(1,:));
  YY1sc     = h*xx(1,BARS(2,:));
  YY2sc     = h*xx(2,BARS(2,:));
  ZZ1sc     = 0.5*(XX1sc+YY1sc);
  ZZ2sc     = 0.5*(XX2sc+YY2sc);
  BARS(3,:) = 1 - 0.9*((cos(Ncell*pi*ZZ1sc).*cos(Ncell*pi*ZZ2sc)).^2);
    
elseif strcmp(flag_matrix,'crac')

  % We create a list of bars to be removed:
  n1a = round(0.4*nside1);
  n1b = round(0.6*nside1);
  n2a = round(0.4*nside2);
  n2b = round(0.7*nside2);
  gap = 4;
  
  ii1 = n1a*nside2 + (1:n2a);
  jj1 = ii1 + nside2;
  ii2 = n1a*nside2 + n2a + 1 + nside2*(1:(n1b-n1a));
  jj2 = ii2 - 1;
  ii3 = n1b*nside2 + 1 + (n2a:n2b);
  jj3 = ii3 + nside2;
  ii4 = (n1b-gap)*nside2 + (nside2:(-1):(n2a+gap+2));
  jj4 = ii4-nside2;
  ii5 = (n1a-1)*nside2 + (n2a+gap+1) + nside2*(1:(n1b-n1a-gap));
  jj5 = ii5+1;
  
  ii = [ii1,ii2,ii3,ii4,ii5];
  jj = [jj1,jj2,jj3,jj4,jj5];
  
  ni   = length(ii);
  BARS(3,:) = ones(1,nbars);
  BARS = [BARS,[ii;jj;-ones(1,ni)]];
      
end

if (flag_anchor == 0)
    
  A = sparse([BARS(1,:), BARS(1,:), BARS(2,:), BARS(2,:)],...
             [BARS(1,:), BARS(2,:), BARS(1,:), BARS(2,:)],...
             [BARS(3,:),-BARS(3,:),-BARS(3,:), BARS(3,:)]);
         
else
    
  ii_rim = [1:nside1,...
            (nside1+1):nside1:(nside1*(nside2-2)+1),...
            (2*nside1):nside1:(nside1*(nside2-1)),...
             nside1*(nside2-1) + (1:nside1)];
     
  A = sparse([BARS(1,:), BARS(1,:), BARS(2,:), BARS(2,:),ii_rim],...
             [BARS(1,:), BARS(2,:), BARS(1,:), BARS(2,:),ii_rim],...
             [BARS(3,:),-BARS(3,:),-BARS(3,:), BARS(3,:),ones(1,2*(nside1+nside2-2))]);
    
end

if strcmp(flag_matrix,'crac')
  [ii,jj,aa] = find(A);
  ind = find(abs(aa) < 1e-12);
  aa(ind) = zeros(size(ind));
  A = sparse(ii,jj,aa);
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This supplementary functions returns a list of all links in a
% rectangular grid of size n1 x n2.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function BARS = LOCAL_get_links(n1,n2)

if (nargin == 1)
  n2 = n1;
end

BARS  = zeros(3,2*n1*n2 - n1 - n2);

ndone = 0;
for i1 = 1:(n1-1)
  for i2 = 1:(n2-1)
    j_c = n2*(i1-1) + i2;
    j_n = n2*(i1-1) + i2 + 1;
    j_e = n2*i1 + i2;
    BARS(1:2,ndone + (1:2)) = [j_c, j_c; j_n, j_e];
    ndone = ndone + 2;
  end
  j_c = n2*i1;
  j_e = n2*(i1+1);
  BARS(1:2,ndone + 1) = [j_c; j_e];
  ndone = ndone + 1;
end
for i2 = 1:(n2-1)
  j_c = n2*(n1-1) + i2;
  j_n = n2*(n1-1) + i2 + 1;
  BARS(1:2,ndone + 1) = [j_c; j_n];
  ndone = ndone + 1;
end

BARS(3,:) = ones(1,ndone);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% NODES{02,ibox} = ilevel
% NODES{03,ibox} = ifather
% NODES{04,ibox} = ison_sw
% NODES{05,ibox} = ison_nw
% NODES{06,ibox} = ison_se
% NODES{07,ibox} = ison_ne
% NODES{09,ibox} = nside   (number of nodes on the side of the box)
% NODES{10,ibox} = i1 coordinate of the box
% NODES{11,ibox} = i2 coordinate of the box
% NODES{12,ibox} = 
% NODES{13,ibox} = C_top      Cholesky factorization of "A22" (computed only on occasion for the top box)
% NODES{14,ibox} = S_inv      inverse of Schur complement     (computed only on occasion for the top box)
% NODES{15,ibox} = S          Schur complement for the box
% NODES{16,ibox} = A_int_inv  inverse of the "interior" matrix  "inv(A11)"
% NODES{17,ibox} = F          projection matrix                 "-inv(A11)*A12"
% NODES{18,ibox} = G          elongation matrix                 "-A21*inv(A11)"
% NODES{19,ibox} = C          Cholesky factorization of A11
% NODES{20,ibox} = H          Cholesky projection matrix        "inv(C)*A12"
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function NODES = LOCAL_get_Ainv_hierarchical(A,nlevel,nleaf)

% Construct the data structure.
NODES = LOCAL_get_tree(nlevel,nleaf);

% Construct the list specifying the order in which the boxes are to be
% scanned. While following this list, any box encountered will either be 
% a leaf box, or a box whose children have already been processed.
box_list = LOCAL_scan_tree(NODES);

% The actual loop.
for icount = 1:length(box_list)
    
  ibox = box_list(icount);

  if (NODES{2,ibox} == nlevel) % ibox is a leaf
    
    [S,Dinv,F,G]  = LOCAL_process_leaf(A,NODES,ibox);
    NODES{15,ibox} = S;
    NODES{16,ibox} = Dinv;
    NODES{17,ibox} = F;
    NODES{18,ibox} = G;

  else % ibox is not a leaf
    
    [S,Dinv,F,G]  = LOCAL_merge_four(NODES,A,ibox);
    NODES{15,ibox} = S;
    NODES{16,ibox} = Dinv;
    NODES{17,ibox} = F;
    NODES{18,ibox} = G;
    NODES{15,NODES{4,ibox}} = [];
    NODES{15,NODES{5,ibox}} = [];
    NODES{15,NODES{6,ibox}} = [];
    NODES{15,NODES{7,ibox}} = [];

  end
  
end

NODES{14,1} = inv(NODES{15,1});

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [S,Dinv,F,G] = LOCAL_process_leaf(A,NODES,ibox)
   
[J_int,J_rim] = LOCAL_get_J_int_J_rim(NODES,ibox);
Dinv          = inv(full(A(J_int,J_int)));
F             = -full(A(J_rim,J_int)*Dinv);
G             = -full(Dinv*A(J_int,J_rim));
S             = full(A(J_rim,J_rim) - A(J_rim,J_int)*Dinv*A(J_int,J_rim));

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [S,Dinv,F,G]  = LOCAL_merge_four(NODES,A,ibox)

ison_sw = NODES{4,ibox};
ison_nw = NODES{5,ibox};
ison_se = NODES{6,ibox};
ison_ne = NODES{7,ibox};

nside   = NODES{9,ison_sw};

[J_sw_sw,J_sw_s,J_sw_se,J_sw_e,J_sw_ne,J_sw_n,J_sw_nw,J_sw_w] = LOCAL_get_J_pieces(NODES,ison_sw);
[J_nw_sw,J_nw_s,J_nw_se,J_nw_e,J_nw_ne,J_nw_n,J_nw_nw,J_nw_w] = LOCAL_get_J_pieces(NODES,ison_nw);
[J_se_sw,J_se_s,J_se_se,J_se_e,J_se_ne,J_se_n,J_se_nw,J_se_w] = LOCAL_get_J_pieces(NODES,ison_se);
[J_ne_sw,J_ne_s,J_ne_se,J_ne_e,J_ne_ne,J_ne_n,J_ne_nw,J_ne_w] = LOCAL_get_J_pieces(NODES,ison_ne);
    
J_1  = [J_sw_e,  J_sw_ne, J_sw_n];
J_2  = [J_se_n,  J_se_nw, J_se_w];
J_3  = [J_ne_w,  J_ne_sw, J_ne_s];
J_4  = [J_nw_s,  J_nw_se, J_nw_e];

J_5  = [J_sw_sw, J_sw_s,  J_sw_se];
J_6  = [J_se_sw, J_se_s,  J_se_se, J_se_e,  J_se_ne];
J_7  = [J_ne_se, J_ne_e,  J_ne_ne, J_ne_n,  J_ne_nw];
J_8  = [J_nw_ne, J_nw_n,  J_nw_nw, J_nw_w,  J_nw_sw];
J_9  = [J_sw_nw, J_sw_w];

I_1  = (nside+1):(3*nside-3);
I_2  = (2*nside):(4*nside-4);
I_3  = [(3*nside-1):(4*nside-4),1:(nside-1)];
I_4  = 2:(2*nside-2);

I_5  = 1:nside;
I_6  = 1:(2*nside-1);
I_7  = nside:(3*nside-2);
I_8  = [(2*nside-1):(4*nside-4),1];
I_9  = (3*nside-2):(4*nside-4);

S_sw = NODES{15,ison_sw};
S_nw = NODES{15,ison_nw};
S_se = NODES{15,ison_se};
S_ne = NODES{15,ison_ne};

X11  = [S_sw(I_1,I_1), A(   J_1,J_2), A(   J_1,J_3), A(   J_1,J_4);...
        A(   J_2,J_1), S_se(I_2,I_2), A(   J_2,J_3), A(   J_2,J_4);... 
        A(   J_3,J_1), A(   J_3,J_2), S_ne(I_3,I_3), A(   J_3,J_4);... 
        A(   J_4,J_1), A(   J_4,J_2), A(   J_4,J_3), S_nw(I_4,I_4)];

X12  = [S_sw(I_1,I_5), A(   J_1,J_6), A(   J_1,J_7), A(   J_1,J_8), S_sw(I_1,I_9);...
        A(   J_2,J_5), S_se(I_2,I_6), A(   J_2,J_7), A(   J_2,J_8), A(   J_2,J_9);... 
        A(   J_3,J_5), A(   J_3,J_6), S_ne(I_3,I_7), A(   J_3,J_8), A(   J_3,J_9);... 
        A(   J_4,J_5), A(   J_4,J_6), A(   J_4,J_7), S_nw(I_4,I_8), A(   J_4,J_9)];

X21  = [S_sw(I_5,I_1), A(   J_5,J_2), A(   J_5,J_3), A(   J_5,J_4);...
        A(   J_6,J_1), S_se(I_6,I_2), A(   J_6,J_3), A(   J_6,J_4);... 
        A(   J_7,J_1), A(   J_7,J_2), S_ne(I_7,I_3), A(   J_7,J_4);... 
        A(   J_8,J_1), A(   J_8,J_2), A(   J_8,J_3), S_nw(I_8,I_4);...
        S_sw(I_9,I_1), A(   J_9,J_2), A(   J_9,J_3), A(   J_9,J_4)];

X22  = [S_sw(I_5,I_5), A(   J_5,J_6), A(   J_5,J_7), A(   J_5,J_8), S_sw(I_5,I_9);...
        A(   J_6,J_5), S_se(I_6,I_6), A(   J_6,J_7), A(   J_6,J_8), A(   J_6,J_9);... 
        A(   J_7,J_5), A(   J_7,J_6), S_ne(I_7,I_7), A(   J_7,J_8), A(   J_7,J_9);... 
        A(   J_8,J_5), A(   J_8,J_6), A(   J_8,J_7), S_nw(I_8,I_8), A(   J_8,J_9);...
        S_sw(I_9,I_5), A(   J_9,J_6), A(   J_9,J_7), A(   J_9,J_8), S_sw(I_9,I_9)];

Dinv = inv(full(X11));
F    = -full(X21*Dinv);
G    = -full(Dinv*X12);
S    = full(X22 - X21*Dinv*X12);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function NODES = LOCAL_get_tree(nlevel,nleaf)

nside  = nleaf*(2^nlevel);
ncells = (4^(nlevel+1) - 1)/3;
NODES  = cell(20,ncells);
NODES{ 2,1} = 0;
NODES{ 3,1} = NaN;
NODES{ 4,1} = 2;
NODES{ 5,1} = 3;
NODES{ 6,1} = 4;
NODES{ 7,1} = 5;
NODES{ 9,1} = nside;
NODES{10,1} = 1;
NODES{11,1} = 1;
ncreated    = 1;
for ilevel = 0:(nlevel-1)
  for ibox = (4^ilevel-1)/3 + (1:(4^ilevel))
    ison1 = ncreated + 1;
    ison2 = ncreated + 2;
    ison3 = ncreated + 3;
    ison4 = ncreated + 4;
    NODES{ 4,ibox } = ison1;
    NODES{ 5,ibox } = ison2;
    NODES{ 6,ibox } = ison3;
    NODES{ 7,ibox } = ison4;
    NODES{ 2,ison1} = ilevel+1;
    NODES{ 2,ison2} = ilevel+1;
    NODES{ 2,ison3} = ilevel+1;
    NODES{ 2,ison4} = ilevel+1;
    NODES{ 3,ison1} = ibox;
    NODES{ 3,ison2} = ibox;
    NODES{ 3,ison3} = ibox;
    NODES{ 3,ison4} = ibox;
    NODES{ 4,ison1} = -1;
    NODES{ 4,ison2} = -1;
    NODES{ 4,ison3} = -1;
    NODES{ 4,ison4} = -1;
    NODES{ 5,ison1} = -1;
    NODES{ 5,ison2} = -1;
    NODES{ 5,ison3} = -1;
    NODES{ 5,ison4} = -1;
    NODES{ 6,ison1} = -1;
    NODES{ 6,ison2} = -1;
    NODES{ 6,ison3} = -1;
    NODES{ 6,ison4} = -1;
    NODES{ 7,ison1} = -1;
    NODES{ 7,ison2} = -1;
    NODES{ 7,ison3} = -1;
    NODES{ 8,ison4} = -1;
    NODES{ 9,ison1} = nleaf*(2^(nlevel-ilevel-1));
    NODES{ 9,ison2} = nleaf*(2^(nlevel-ilevel-1));
    NODES{ 9,ison3} = nleaf*(2^(nlevel-ilevel-1));
    NODES{ 9,ison4} = nleaf*(2^(nlevel-ilevel-1));
    NODES{10,ison1} = 2*NODES{10,ibox}-1;
    NODES{11,ison1} = 2*NODES{11,ibox}-1;
    NODES{10,ison2} = 2*NODES{10,ibox}-1;
    NODES{11,ison2} = 2*NODES{11,ibox}  ;
    NODES{10,ison3} = 2*NODES{10,ibox}  ;
    NODES{11,ison3} = 2*NODES{11,ibox}-1;
    NODES{10,ison4} = 2*NODES{10,ibox}  ;
    NODES{11,ison4} = 2*NODES{11,ibox}  ;
    ncreated = ncreated + 4;
  end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [J_int,J_rim] = LOCAL_get_J_int_J_rim(NODES,ibox)

ilevel = NODES{ 2,ibox};
ncell  = NODES{ 9,ibox};
i1     = NODES{10,ibox};
i2     = NODES{11,ibox};
ncol   = ncell * (2^ilevel);
ifirst = ncol*ncell*(i1-1) + ncell*(i2-1) + 1;

J_int  = zeros(1,(ncell-2)*(ncell-2));
for i = 1:(ncell-2)
  J_int((i-1)*(ncell-2) + (1:(ncell-2))) = ...
      ncol*(ncell*(i1-1)+i) + ncell*(i2-1) + (2:(ncell-1));
end

J_rim = [ifirst + ncol*(0:(ncell-1)),... 
         ifirst + ncol*(ncell-1) + (1:(ncell-2)),...
         ncol*(i1*ncell - 1) + ncell*i2 + ncol*(0:(-1):(-ncell+1)),...
         ifirst + ((ncell-2):(-1):1)];

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function box_list = LOCAL_scan_tree(NODES)
box_list          = zeros(1,size(NODES,2));
nfound            = 0;
[box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,1);
return
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,ibox)
ison1 = NODES{4,ibox};
ison2 = NODES{5,ibox};
ison3 = NODES{6,ibox};
ison4 = NODES{7,ibox};
if (ison1 > -1)
  [box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,ison1);
end
if (ison2 > -1)
  [box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,ison2);
end
if (ison3 > -1)
  [box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,ison3);
end
if (ison4 > -1)
  [box_list,nfound] = LOCAL_scan_tree_rec(NODES,box_list,nfound,ison4);
end
box_list(nfound+1) = ibox;
nfound             = nfound+1;
return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Jsw,Js,Jse,Je,Jne,Jn,Jnw,Jw] = LOCAL_get_J_pieces(NODES,ibox)

ilevel = NODES{ 2,ibox};
ncell  = NODES{ 9,ibox};
i1     = NODES{10,ibox};
i2     = NODES{11,ibox};

ncol   = ncell * (2^ilevel);
ifirst = ncol*ncell*(i1-1) + ncell*(i2-1) + 1;

J_rim  = [ifirst + ncol*(0:(ncell-1)),...
          ifirst + ncol*(ncell-1) + (1:(ncell-1)),...
          ncol*(i1*ncell - 1) + ncell*i2 + ncol*((-1):(-1):(-ncell+1)),...
          ifirst + ((ncell-2):(-1):1)];

Jsw = ifirst;
Js  = ifirst + ncol*(1:(ncell-2));
Jse = ifirst + ncol*(ncell-1);
Je  = ifirst + ncol*(ncell-1) + (1:(ncell-2));
Jne = ifirst + ncol*(ncell-1) + ncell - 1;
Jn  = ncol*(i1*ncell - 1) + ncell*i2 + ncol*((-1):(-1):(-ncell+2));
Jnw = ifirst + ncell - 1;
Jw  = ifirst + ((ncell-2):(-1):1);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Jrim,Jint] = LOCAL_get_Jrim_Jint_for_square(nside)

Jrim = [1 + nside*(0:(nside-1)),...
        1 + nside*(nside-1) + (1:(nside-1)),...
        nside*(nside-1):(-nside):(2*nside),...
        nside:(-1):2];
    
Jint = zeros(1,(nside-2)*(nside-2));

for i = 1:(nside-2)
  Jint((nside-2)*(i-1) + (1:(nside-2))) = 1 + i*nside + (1:(nside-2));
end
      
return
      
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function mem = LOCAL_get_bytes(X)

tmp = whos('X');

mem = tmp.bytes;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function u = LOCAL_apply_Ainv(NODES,f)

u = zeros(size(f));

nboxes = size(NODES,2);

%%% Upwards pass
for ibox = nboxes:(-1):1
  if (NODES{4,ibox} == (-1)) % ibox is a leaf
    [J_int,J_rim] = LOCAL_get_J_int_J_rim(NODES,ibox);
    f(J_rim)      = f(J_rim) + NODES{17,ibox}*f(J_int,:);
  else
    [J_int,J_rim] = LOCAL_get_J_int_J_rim_for_quad(NODES,ibox);
    f(J_rim)      = f(J_rim) + NODES{17,ibox}*f(J_int);
  end
end

%%% Toplevel solve.
[trash,J_rim] = LOCAL_get_J_int_J_rim(NODES,1);
u(J_rim,:) = NODES{14,1}*f(J_rim,:);
 
%%% Downwards pass.
for ibox = 1:nboxes
  if (NODES{4,ibox} == (-1)) % ibox is a leaf
    [J_int,J_rim] = LOCAL_get_J_int_J_rim(NODES,ibox);
    u(J_int,:)    = NODES{16,ibox}*f(J_int,:) + NODES{18,ibox}*u(J_rim,:);
  else
    [J_int,J_rim] = LOCAL_get_J_int_J_rim_for_quad(NODES,ibox);
    u(J_int,:)    = NODES{16,ibox}*f(J_int,:) + NODES{18,ibox}*u(J_rim,:);
  end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [J_int,J_rim] = LOCAL_get_J_int_J_rim_for_quad(NODES,ibox)
ison_sw = NODES{4,ibox};
ison_nw = NODES{5,ibox};
ison_se = NODES{6,ibox};
ison_ne = NODES{7,ibox};
[J_sw_sw,J_sw_s,J_sw_se,J_sw_e,J_sw_ne,J_sw_n,J_sw_nw,J_sw_w] = LOCAL_get_J_pieces(NODES,ison_sw);
[J_nw_sw,J_nw_s,J_nw_se,J_nw_e,J_nw_ne,J_nw_n,J_nw_nw,J_nw_w] = LOCAL_get_J_pieces(NODES,ison_nw);
[J_se_sw,J_se_s,J_se_se,J_se_e,J_se_ne,J_se_n,J_se_nw,J_se_w] = LOCAL_get_J_pieces(NODES,ison_se);
[J_ne_sw,J_ne_s,J_ne_se,J_ne_e,J_ne_ne,J_ne_n,J_ne_nw,J_ne_w] = LOCAL_get_J_pieces(NODES,ison_ne);
J_int = [J_sw_e,J_sw_ne,J_sw_n,...
         J_se_n,J_se_nw,J_se_w,...
         J_ne_w,J_ne_sw,J_ne_s,...
         J_nw_s,J_nw_se,J_nw_e];
J_rim = [J_sw_sw,J_sw_s,J_sw_se,J_se_sw,J_se_s,...
         J_se_se,J_se_e,J_se_ne,J_ne_se,J_ne_e,...
         J_ne_ne,J_ne_n,J_ne_nw,J_nw_ne,J_nw_n,...
         J_nw_nw,J_nw_w,J_nw_sw,J_sw_nw,J_sw_w];
return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
